from SimulatorFramework.TileType import TileType
from SimulatorFramework import Frontier
from SimulatorFramework import SimUtils
import wx
import numpy as np
import random
import copy
import xml.etree.ElementTree as ET
import xml.dom.minidom


DIR_OFFSETS = {
    'N':  np.array([ 0,  1,  0]),
    'S':  np.array([ 0, -1,  0]),
    'E':  np.array([ 1,  0,  0]),
    'W':  np.array([-1,  0,  0]),
    'U':  np.array([ 0,  0,  1]),
    'D':  np.array([ 0,  0, -1])
}

REV_DIR_OFFSETS = {
    ( 0,  1,  0):   'N',
    ( 0, -1,  0):   'S',
    ( 1,  0,  0):   'E',
    (-1,  0,  0):   'W',
    ( 0,  0,  1):   'U',
    ( 0,  0, -1):   'D',
}

DIR_OPPS = {
    'N': 'S',
    'S': 'N',
    'E': 'W',
    'W': 'E',
    'U': 'D',
    'D': 'U',
}

glue_offset = {'N':  (1,0),
               'E':  (2,1),
               'S':  (1,2),
               'W':  (0,1),
               'U':  (1,1),
               'D':  (1,1),
               'NE': (2,0),
               'SE': (2,2),
               'SW': (0,2),
               'NW': (0,0),
}

def make_3D(coords):
    ret_coords = coords
    if len(ret_coords) < 3:
        ret_coords = (coords[0], coords[1], 0)
    return ret_coords

def labels_complement(l1, l2):
    return l1 == l2 + '*' or l2 == l1 + '*'

class PolyominoType:
    def __init__(self, name, color=None, blocks=None, domains=None, xml_node=None):
        self.name = name
        self.color = color
        if xml_node is None:
            self.blocks = blocks or [(0, 0, 0)]
            self.domains = domains or {}
        else:
            self.fromXML(xml_node)
        self.init_color()
        self.ensure_3D()

    def __str__(self):
        return f'Polyomino type {self.name}: blocks: {self.blocks}, domains: {self.domains}'
    
    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if self.name != other.name:
            return False
        if self.domains.keys() != other.domains.keys():
            return False
        for loc in self.domains:
            if set(self.domains[loc]) != set(other.domains[loc]):
                return False
        if self.blocks != other.blocks:
            return False
        return True

    def __hash__(self):
        n = hash(self.name)
        b = hash(tuple(sorted(list(self.blocks))))
        d = hash(tuple(sorted([(l, tuple(sorted(d))) for l, d in self.domains.items()])))
        return hash((n, b, d))
    
    def ensure_3D(self):
        new_blocks = []
        for block in self.blocks:
            new_blocks.append(make_3D(block))
        self.blocks = new_blocks
        new_domains = dict()
        for domain_coords in self.domains:
            new_domains[make_3D(domain_coords)] = self.domains[domain_coords]
        self.domains = new_domains

    def init_color(self):
        if self.color is None:
            self.color = "LIGHT GREY"
        else:
            wx_color = wx.TheColourDatabase.Find(self.color)
            if wx_color == False:
                msg = f"Invalid color '{self.color}' specified for polyomino type {self.name}"
                SimUtils.globalApp.OutputMessage(msg)
                self.color = "LIGHT GREY"
        color = wx.TheColourDatabase.Find(self.color)
        self.colorVector = (color.Red(), color.Green(), color.Blue())

    def Copy(self):
        pct = PolyominoType(name=self.name, color=self.color)
        pct.blocks = copy.deepcopy(self.blocks)
        pct.domains = copy.deepcopy(self.domains)
        return pct
        
    def Draw(self, dc, zero_coords, side_len, scale_factor):
        for block in self.blocks:
            dc.SetBrush(wx.Brush(wx.TheColourDatabase.Find(self.color)))
            self.DrawBlock(dc, zero_coords, side_len, scale_factor, block, self.domains)
            
    def DrawBlock(self, dc, zero_coords, side_len, scale_factor, block, domains):
        x_start = zero_coords[0] + ((side_len * block[0]) * scale_factor)
        y_start = zero_coords[1] - ((side_len * block[1]) * scale_factor)
        
        size = side_len * scale_factor

        top_left = wx.Point(x_start, y_start)
        bottom_right = wx.Point(x_start + size, y_start + size)
        dc.DrawRectangle(wx.Rect(top_left, bottom_right))

        if block in domains:
            dc.SetBrush(wx.Brush(wx.TheColourDatabase.Find("RED")))
            third = int((side_len * scale_factor)/3)
            domain_list = domains[block]
            for domain in domain_list:
                label = domain[0]
                direction = domain[1]
                top_left = wx.Point(x_start+(glue_offset[direction][0]*third), y_start+(glue_offset[direction][1]*third))
                bottom_right = wx.Point(top_left[0]+third, top_left[1]+third)
                dc.DrawRectangle(wx.Rect(top_left, bottom_right))
                dc.DrawLabel(label, wx.Rect(top_left, bottom_right), alignment=wx.ALIGN_CENTER_HORIZONTAL|wx.ALIGN_CENTER_VERTICAL)

    def toXML(self, xml_parent=None):
        if xml_parent is None:
            xml_parent = ET.Element('PolyominoType')
            elt_ptype = xml_parent
        else:
            elt_ptype = ET.SubElement(xml_parent, 'PolyominoType')

        if not self.name is None:
            elt_name = ET.SubElement(elt_ptype, 'name')
            elt_name.text = self.name
        if not self.color is None:
            elt_color = ET.SubElement(elt_ptype, 'color')
            elt_color.text = self.color
            
        elt_blocks = ET.SubElement(elt_ptype, 'blocks')
        for block in self.blocks:
            elt_block = ET.SubElement(elt_blocks, 'block')
            elt_coords = ET.SubElement(elt_block, 'coords')
            elt_coords.text = str(block)
            if block in self.domains:
                elt_domains = ET.SubElement(elt_block, 'domains')
                domain_list = self.domains[block]
                for domain in domain_list:
                    elt_domain = ET.SubElement(elt_domains, 'domain')
                    elt_label = ET.SubElement(elt_domain, 'label')
                    elt_label.text = domain[0]
                    elt_direction = ET.SubElement(elt_domain, 'direction')
                    elt_direction.text = domain[1]
        return xml_parent
    
    def fromXML(self, xml_parent):
        if not xml_parent.tag == 'PolyominoType':
            print("Invalid XML node for PolyominoType")
        else:
            blocks = []
            self.domains = dict()
            for elt_child in xml_parent:
                if elt_child.tag == 'name':
                    self.name = elt_child.text
                elif elt_child.tag == 'color':
                    self.color = elt_child.text
                elif elt_child.tag == 'blocks':
                    # get blocks
                    for elt_block in elt_child:
                        if elt_block.tag == 'block':
                            for elt_child2 in elt_block:
                                if elt_child2.tag == 'coords':
                                    coords = eval(elt_child2.text)
                                    blocks.append(coords)
                                elif elt_child2.tag == 'domains':
                                    # get domains
                                    self.domains[coords] = []
                                    for elt_domain in elt_child2:
                                        if elt_domain.tag == 'domain':
                                            label = None
                                            direction = None
                                            for elt_child3 in elt_domain:
                                                if elt_child3.tag == 'label':
                                                    label = elt_child3.text
                                                elif elt_child3.tag == 'direction':
                                                    direction = elt_child3.text
                                            if (not label is None) and (not coords is None):
                                                self.domains[coords].append((label, direction))
            self.blocks = blocks

    def getBoundingBox(self, blocks=None):
        min_x, min_y, min_z =  999999, 999999,  999999
        max_x, max_y, max_z = -999999,-999999, -999999
        bEmpty = True
        
        if blocks is None:
            blocks = self.blocks

        for block in blocks:
            bEmpty = False
            if block[0] < min_x:
                min_x = block[0]
            if block[0] > max_x:
                max_x = block[0]
            if block[1] < min_y:
                min_y = block[1]
            if block[1] > max_y:
                max_y = block[1]
            if block[2] < min_z:
                min_z = block[2]
            if block[2] > max_z:
                max_z = block[2]

        if bEmpty == True:
            return ((0,0), (0,0), (0,0))
        else:
            return ((min_x, min_y, min_z), (max_x, max_y, max_z))

            
class Polyomino:
    def __init__(self, polyominoType, xml_node=None, polyominoTypes=None):
        if xml_node is None:
            self.polyominoType = polyominoType
            self.color = self.polyominoType.color
            self.blocks = copy.deepcopy(self.polyominoType.blocks)
            self.domains = copy.deepcopy(self.polyominoType.domains)
        else:
            self.fromXML(xml_node, polyominoTypes)
        self.ensure_3D()
     
    def __str__(self):
        return f'Polyomino of type {self.polyominoType.name}: blocks: {self.blocks}, domains: {self.domains}'

    def __eq__(self, other):
        if self.polyominoType != other.polyominoType:
            return False
        if self.domains.keys() != other.domains.keys():
            return False
        for loc in self.domains:
            if set(self.domains[loc]) != set(other.domains[loc]):
                return False
        if self.blocks != other.blocks:
            return False
        return True

    def ensure_3D(self):
        new_blocks = []
        for block in self.blocks:
            new_blocks.append(make_3D(block))
        self.blocks = new_blocks
        new_domains = dict()
        for domain_coords in self.domains:
            new_domains[make_3D(domain_coords)] = self.domains[domain_coords]
        self.domains = new_domains
                
    def __hash__(self):
        n = hash(self.polyominoType.name)
        b = hash(tuple(sorted(list(self.blocks))))
        d = hash(tuple(sorted([(l, tuple(sorted(d))) for l, d in self.domains.items()])))
        return hash((n, b, d))

    def translate(self, vector):
        if len(vector) < 3:
            vector = (vector[0], vector[1], 0)
        new_blocks = []
        for block in self.blocks:
            x,y,z = block[0] + vector[0], block[1] + vector[1], block[2] + vector[2]
            new_blocks.append((x,y,z))
        self.blocks = new_blocks

        translated_domains = dict()
        for domain in self.domains:
            loc = (domain[0] + vector[0], domain[1] + vector[1], domain[2] + vector[2])
            translated_domains[loc] = self.domains[domain]
        self.domains = translated_domains

    def Draw(self, dc, zero_coords, side_len, scale_factor):
        for block in self.blocks:
            dc.SetBrush(wx.Brush(wx.TheColourDatabase.Find(self.color)))
            self.polyominoType.DrawBlock(dc, zero_coords, side_len, scale_factor, block, self.domains)

    def getTranslation(self):
        bError = False
        diff2 = None
        for bcount in range(len(self.blocks)):
            diff = (self.blocks[bcount][0]-self.polyominoType.blocks[bcount][0],
                    self.blocks[bcount][1]-self.polyominoType.blocks[bcount][1],
                    self.blocks[bcount][2]-self.polyominoType.blocks[bcount][2])
            if (not diff2 is None) and (not diff == diff2):
                bError = True
            diff2 = diff
        if bError:
            print("ERROR: polyomino translation is not consistent")
        return diff
                                                                               
    def getBoundingBox(self):
        return self.polyominoType.getBoundingBox(self.blocks)

    def toXML(self, xml_parent=None):
        if xml_parent is None:
            xml_parent = ET.Element('Polyomino')
            elt_polyomino = xml_parent
        else:
            elt_polyomino = ET.SubElement(xml_parent, 'Polyomino')

        elt_polyominotype = ET.SubElement(elt_polyomino, 'PolyominoType')
        elt_polyominotype.text = self.polyominoType.name

        elt_translation = ET.SubElement(elt_polyomino, 'translation')
        elt_translation.text = str(self.getTranslation())

        return xml_parent

    def fromXML(self, xml_parent, polyominoTypes):
        if not xml_parent.tag == 'Polyomino':
            print("Invalid XML node for Polyomino")
        else:
            self.polyominoType = PolyominoType(None)
            blocks = []
            self.domains = dict()
            translation = (0,0,0)
            for elt_child in xml_parent:
                if elt_child.tag == 'PolyominoType':
                    polyominoType_name = elt_child.text
                    if not polyominoType_name in polyominoTypes:
                        print(f"Error loading polyomino: Invalid polyomino type '{polyominoType_name}")
                        self.polyominoType = None
                    else:
                        self.polyominoType = polyominoTypes[polyominoType_name]
                elif elt_child.tag == 'translation':
                    translation = eval(elt_child.text)
         
        if not self.polyominoType is None:
            self.color = self.polyominoType.color
            self.blocks = copy.deepcopy(self.polyominoType.blocks)
            self.domains = copy.deepcopy(self.polyominoType.domains)
            self.translate(translation)
